from __future__ import annotations

import json
import time

from pathlib import Path

from b22ao.aosystem import AOSystem
from b22ao.message import Message, State

from typing import Optional

import logging
logger = logging.getLogger('b22ao.base')

class BaseOperation:

    """
    Base class for any adaptive optics routine.
    Children implement #start and #stop, and can #deform the mirrors and #capture data from the camera at their leisure!
    """
   
    def setup_ao_system(self):
        self.ao = AOSystem()
        logger.debug('AO system initialised')

    def attach_listener(self, listener):
        self.listener = listener

    def load_config(self, config: Optional[str | Path]):
        if config:
            with open(config, 'r') as doc:
                self.config = json.load(doc)
        else:
            self.config = {}

        logger.debug(f'Loaded the following configuration: {self.config}')

    def select_dm(self, dm: int):
        """
        If your entire operation involves a single mirror, you can specify it here

        e.g. passed in as config:

        self.select_dm(self.config['mirror'])  # where "mirror": 2 in config file
        # later:
        self.deform(mask)  # no need to specify mirror in subsequent calls
        """
        self.ao.select_dm(dm)

    def deform(self, mask, mirror: Optional[int] = None):
        """
        Send a mask to specified mirror (1 or 2).

        If a mirror was previously selected with #select_dm
        and no mirror is specified here, the mask is applied
        to the mirror selected earlier.

        Bear in mind that the deformable mirrors in B22 *do not* support readback,
        so there is no guarantee that the mask is fully applied by the time this
        function returns; it may be necessary to wait for an empirically-determined
        time before characterising the system.
        """
        self.ao.deform(mask, mirror)

    def capture(self):
        """
        Returns a single detector frame
        """
        return self.ao.capture()
    
    def deform_and_capture(self, mask, mirror: Optional[int] = None, delay_s: float = 0.01):
        """
        Applies mask and returns single detector frame after small delay
        """
        self.deform(mask, mirror)
        time.sleep(delay_s)
        return self.capture()

    def run(self):
        """
        Do not override. Implement #start instead
        """
        logger.info('Starting operation')
        self.notify(Message(self, State.Running))
        self.start()
        self.notify(Message(self, State.Idle))
        logger.info('Operation ended')

    def abort(self):
        """
        Do not override. Implement #stop instead
        """
        logger.info('Aborting operation')
        self.stop()
        self.notify(Message(self, State.Idle, "Aborted"))

    def start(self):
        """
        Starts the operation

        If a JSON configuration file was specified,
        implementations can now access the dictionary self.config
        """
        raise NotImplementedError

    def stop(self):
        """
        Stops the operation
        """
        raise NotImplementedError

    def report_progress(self, iteration, value, **other_stuff):
        """
        Inform a listening component (e.g. a live plot) of the latest result.
        """
        msg = dict({"iteration": iteration, "value": value}, **other_stuff)
        self.notify(Message(self, State.Running, msg))

    def report_error(self, error_msg: str):
        msg = {"error": error_msg}
        self.notify(Message(self, State.Error, msg))

    def notify(self, message: Message):
        logger.debug(f'Notifying listener: {message}')
        try:
            self.listener.notify(message)
        except AttributeError:
            print(message)



import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
stream_handler.setFormatter(logging.Formatter('%(message)s'))

logger.addHandler(stream_handler)

def configure(filepath=None, graylog_host=None, graylog_port=None):
    
    formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s || %(message)s",
                datefmt="%d-%m-%Y %I:%M:%S")

    def setup_file_handler(file_path):
        file_handler = logging.FileHandler(file_path, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    
    def setup_graylog_handler(host, port):
        from graypy import GELFUDPHandler
        graylog_handler = GELFUDPHandler(host, port)
        graylog_handler.setFormatter(formatter)
        logger.addHandler(graylog_handler)
    
    if filepath:
        setup_file_handler(filepath)
    
    if graylog_host and graylog_port:
        setup_graylog_handler(graylog_host, graylog_port)



from epics import caget, caput, camonitor
from b22ao.pvs import *
from enum import Enum
from threading import Event
from p4p.client.thread import Context
import numpy
import logging
import time

logger = logging.getLogger('b22ao.aosystem')

class AOSystem:

    def __init__(self):
        self.cam = AreaDetector(AD_PVBASE)
        self.dm = None  # DM of choice

    def select_dm(self, dm):
        logger.info(f'DM {dm} selected')
        self.dm = DeformableMirror(dm)

    def deform(self, mask, dm=None):
        try:
            DeformableMirror(dm).deform(mask)
        except ValueError:  # no dm provided
            try:
                self.dm.deform(mask)
            except AttributeError:
                raise NameError("Call #select_dm(dm) first or specify a DM in #deform(mask, dm)")

    def capture(self):
        return self.cam.acquire()

    def get_metadata(self):
        return {'cam': self.cam.get_metadata(), 'dm': self.dm.get_metadata()}


class DeformableMirror(Enum):

    DM1 = 1
    DM2 = 2

    def deform(self, mask):
        logger.debug('Deforming mirror')
        for actuator in range(len(mask)):
            caput(self._get_pv_base() + DM_ACTUATOR_PREFIX + str(actuator) + DM_ACTUATOR_SETPOINT, mask[actuator])
        caput(self._get_pv_base() + DM_APPLY_MASK, 1)

    def _get_pv_base(self):
        if self is DeformableMirror.DM1:
            return standardise_pv(DM1_PVBASE)
        elif self is DeformableMirror.DM2:
            return standardise_pv(DM2_PVBASE)

    def get_metadata(self):
        return {'mirror': self}


class AreaDetector:

    def __init__(self, pv_base):
        pv_base = standardise_pv(pv_base)
        self.acquisition_allowed = Event()
        self.acquisition_complete = Event()
        self.data_ready = Event()
        self.error_state = False
        self.current_frame = None
        self.frame_counter_pv = pv_base + AD_ARRAY_COUNTER
        self.acquire_pv = pv_base + AD_ACQUIRE
        self.data_pv = pv_base + AD_ARRAY_DATA
        self.gain_pv = pv_base + AD_GAIN
        self.shutter_pv = pv_base + AD_SHUTTER_STATE
<<<<<<< HEAD
=======
        self.state_pv = pv_base + AD_STATE
>>>>>>> error_recovery

        caput(pv_base + AD_IMAGE_MODE, AD_IMAGE_MODE_SINGLE)

        camonitor(self.state_pv, callback=self.state_callback)
        camonitor(self.frame_counter_pv, callback=self.frame_counter_callback)
        camonitor(ALLOW_ACQUISITION, callback=self.allow_acquisition_callback)
        
        # synchronise acquisition_allowed event with current PV value
        self.allow_acquisition_callback(value=caget(ALLOW_ACQUISITION))

        self.pva_ctxt = Context('pva')

    def acquire(self):
        """
        Returns a valid frame
        (acquires and reads shutter state:
        repeats until shutter is open)
        """
        # request frames until shutter is open
        while True:
            self.acquisition_allowed.wait()
            self.request_frame()
            if caget(self.shutter_pv, as_string=True) == "Open":
                break
            logger.debug("Frame received, but shutter closed. Reacquiring...")
        
        # shutter is open; read data
        pvadata = self.pva_ctxt.get(self.data_pv)
        data = numpy.reshape(pvadata.data, pvadata.shape)
        logger.debug("Acquisition successful")
        return data

    def request_frame(self, attempt=0, max_attempts=5, time_before_retry=5):
        """
        Hits the 'Acquire' button and blocks
        until data is ready to be read from Array plugin
        """
        self.current_frame = caget(self.frame_counter_pv)
        logger.debug("Acquiring")
        caput(self.acquire_pv, "Acquire")
        self.acquisition_complete.wait()

        if self.error_state:

            if attempt >= max_attempts:
                raise Exception("Max retries reached - giving up")
            
            logger.error(f"IOC in Error state. Retrying acquisition in {time_before_retry} seconds")
            caput(self.acquire_pv, "Done")  # i.e. smash the stop button
            time.sleep(time_before_retry)
            self.request_frame(attempt=attempt+1)
        
        # Hooray: Acquisition done and detector not in error state.
        logger.debug("Waiting for frame to appear in Area Detector")
        self.data_ready.wait()
        self.data_ready.clear()

    def state_callback(self, **kwargs):
        state = kwargs['value']
        if state == 1: # Acquire
            self.acquisition_complete.clear()
        elif state == 0: # Idle
            self.acquisition_complete.set()
            self.error_state = False
        elif state == 6: # Error
            self.acquisition_complete.set()
            self.error_state = True
    
    def frame_counter_callback(self, **kwargs):
        try:
            if kwargs['value'] == self.current_frame + 1:
                self.data_ready.set()
        except AttributeError:
            # callback called before self.current_frame was initialised
            pass

    def allow_acquisition_callback(self, **kwargs):
        value = kwargs['value']
        if value == 0:
            logger.info("Acquisition disallowed")
            self.acquisition_allowed.clear()
        elif value == 1:
            logger.info("Acquisition allowed")
            self.acquisition_allowed.set()

    def get_metadata(self):
        metadata = dict()
        metadata['gain'] = caget(self.gain_pv)
        return metadata


def standardise_pv(pv):
    return pv if pv.endswith(':') else pv + ':'




AD_PVBASE = "BL22B-DI-CAM-10"  # Area Detector

DM1_PVBASE = "BL22B-OP-ALPAO-01"  # Deformable mirror 1
DM2_PVBASE = "BL22B-OP-ALPAO-02"  # Deformable mirror 2


AD_ARRAY_COUNTER = "PVA:UniqueId_RBV"  # Frame counter
AD_ACQUIRE = "CAM:Acquire"  # Acquire button - write with 'Acquire'
AD_GAIN = "CAM:Gain_RBV"  # Gain set on Windows app
AD_SHUTTER_STATE = "CAM:SHUTTER_RBV"
AD_ARRAY_DATA = "PVA:ARRAY" # Structured data - use p4p calls
AD_IMAGE_MODE = "CAM:ImageMode"
AD_IMAGE_MODE_SINGLE = "Single"
AD_STATE = "CAM:DetectorState_RBV"

DM_ACTUATOR_PREFIX = "ACT"  # individual actuator pv: pv base + ACT + actuator number
DM_ACTUATOR_SETPOINT = ":SP"
DM_APPLY_MASK = "CP_ST_TO_ACT.PROC"

'''
The following PV is a gate signal used to allow/inhibit acquisitions.
The signal responds to topup events, beam dump, and a manual override.
PV values: 0 = disallow; 1 = allow

Note that this is not the exact PV accessible via the EPICS screen.
That PV adjusts this result to the specific endstation in use (the
signal is inverted for one of them). The PV in use here is the
endstation-independent result of their gating logic.
'''
ALLOW_ACQUISITION = "BL22B-EA-FTIR-02:CALC"


from enum import Enum
from typing import Any, Dict

class State(Enum):
    Idle = "Idle"
    Running = "Running"
    Error = "Error"


class Message:
    def __init__(self, source: Any, state: State, msg: Dict=None):
        self.source = source
        self.state = state
        self.msg = msg
